@@ -44,7 +44,6 @@
 #include "body.h"
 #include "cell.h"
 #include "image.h"
-#include "link.h"
 #include "menubar.h"
 #include "popup.h"
 #include "properties.h"
@@ -58,22 +57,58 @@
 #include "gtkhtmldebug.h"
 #endif
 
+static void
+undo (GtkWidget *mi, GtkHTMLControlData *cd)
+{
+	gtk_html_undo (cd->html);
+}
+
+static void
+redo (GtkWidget *mi, GtkHTMLControlData *cd)
+{
+	gtk_html_redo (cd->html);
+}
+
 static void
 copy (GtkWidget *mi, GtkHTMLControlData *cd)
 {
-	html_engine_copy (cd->html->engine);
+	gtk_html_copy (cd->html);
 }
 
 static void
 cut (GtkWidget *mi, GtkHTMLControlData *cd)
 {
-	html_engine_cut (cd->html->engine);
+	gtk_html_cut (cd->html);
 }
 
 static void
 paste (GtkWidget *mi, GtkHTMLControlData *cd)
 {
-	html_engine_paste (cd->html->engine);
+	gtk_html_paste (cd->html, FALSE);
+}
+
+static void
+paste_cite (GtkWidget *mi, GtkHTMLControlData *cd)
+{
+	gtk_html_paste (cd->html, TRUE);
+}
+
+static void
+insert_link (GtkWidget *mi, GtkHTMLControlData *cd)
+{
+	if (cd->properties_dialog)
+		gtk_html_edit_properties_dialog_close (cd->properties_dialog);
+
+	cd->properties_dialog = gtk_html_edit_properties_dialog_new (cd, TRUE, _("Insert"), ICONDIR "/insert-link-24.png");
+
+	gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
+						   GTK_HTML_EDIT_PROPERTY_LINK, _("Link"),
+						   link_insert,
+						   link_insert_cb,
+						   link_close_cb);
+
+	gtk_html_edit_properties_dialog_show (cd->properties_dialog);
+	gtk_html_edit_properties_dialog_set_page (cd->properties_dialog, GTK_HTML_EDIT_PROPERTY_LINK);
 }
 
 static void
@@ -82,7 +117,7 @@ remove_link (GtkWidget *mi, GtkHTMLControlData *cd)
 	html_engine_selection_push (cd->html->engine);
 	if (!html_engine_is_selection_active (cd->html->engine))
 		html_engine_select_word_editable (cd->html->engine);
-	html_engine_insert_link (cd->html->engine, NULL, NULL);
+	html_engine_edit_set_link (cd->html->engine, NULL, NULL);
 	html_engine_selection_pop (cd->html->engine);
 }
 
@@ -146,7 +181,9 @@ show_prop_dialog (GtkHTMLControlData *cd, GtkHTMLEditPropertyType start)
 	GtkHTMLEditPropertyType t;
 	GList *cur;
 
-	cd->properties_dialog = gtk_html_edit_properties_dialog_new (cd, FALSE, _("Properties"));
+	if (cd->properties_dialog)
+		gtk_html_edit_properties_dialog_close (cd->properties_dialog);
+	cd->properties_dialog = gtk_html_edit_properties_dialog_new (cd, FALSE, _("Properties"), ICONDIR "/properties-16.png");
 
 	cur = cd->properties_types;
 	while (cur) {
@@ -159,6 +196,13 @@ show_prop_dialog (GtkHTMLControlData *cd, GtkHTMLEditPropertyType start)
 								   text_apply_cb,
 								   text_close_cb);
 			break;
+		case GTK_HTML_EDIT_PROPERTY_LINK:
+			gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
+								   t, _("Link"),
+								   link_properties,
+								   link_apply_cb,
+								   link_close_cb);
+			break;
 		case GTK_HTML_EDIT_PROPERTY_IMAGE:
 			gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
 								   t, _("Image"),
@@ -173,13 +217,6 @@ show_prop_dialog (GtkHTMLControlData *cd, GtkHTMLEditPropertyType start)
 								   paragraph_apply_cb,
 								   paragraph_close_cb);
 			break;
-		case GTK_HTML_EDIT_PROPERTY_LINK:
-			gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
-								   t, _("Link"),
-								   link_properties,
-								   link_apply_cb,
-								   link_close_cb);
-			break;
 		case GTK_HTML_EDIT_PROPERTY_BODY:
 			gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
 								   t, _("Page"),
@@ -225,6 +262,24 @@ prop_dialog (GtkWidget *mi, GtkHTMLControlData *cd)
 	show_prop_dialog (cd, GPOINTER_TO_INT (gtk_object_get_data (GTK_OBJECT (mi), "type")));
 }
 
+static void
+link_prop_dialog (GtkWidget *mi, GtkHTMLControlData *cd)
+{
+	if (cd->properties_dialog)
+		gtk_html_edit_properties_dialog_close (cd->properties_dialog);
+
+	cd->properties_dialog = gtk_html_edit_properties_dialog_new (cd, FALSE, _("Properties"), ICONDIR "/insert-link-24.png");
+
+	gtk_html_edit_properties_dialog_add_entry (cd->properties_dialog,
+						   GTK_HTML_EDIT_PROPERTY_LINK, _("Link"),
+						   link_properties,
+						   link_apply_cb,
+						   link_close_cb);
+
+	gtk_html_edit_properties_dialog_show (cd->properties_dialog);
+	gtk_html_edit_properties_dialog_set_page (cd->properties_dialog, GTK_HTML_EDIT_PROPERTY_LINK);
+}
+
 static void
 spell_suggest (GtkWidget *mi, GtkHTMLControlData *cd)
 {
@@ -232,7 +287,7 @@ spell_suggest (GtkWidget *mi, GtkHTMLControlData *cd)
 
 	/* gtk_signal_emit_by_name (GTK_OBJECT (cd->html), "spell_suggestion_request",
 	   e->spell_checker, html_engine_get_word (e)); */
-	spell_suggestion_request (cd->html, html_engine_get_word (e), cd);
+	spell_suggestion_request (cd->html, html_engine_get_spell_word (e), cd);
 }
 
 static void
@@ -247,7 +302,7 @@ spell_add (GtkWidget *mi, GtkHTMLControlData *cd)
 	HTMLEngine *e = cd->html->engine;
 	gchar *word;
 
-	word = html_engine_get_word (e);
+	word = html_engine_get_spell_word (e);
 	if (word) {
 		spell_add_to_personal (cd->html, word, cd);
 		g_free (word);
@@ -261,7 +316,7 @@ spell_ignore (GtkWidget *mi, GtkHTMLControlData *cd)
 	HTMLEngine *e = cd->html->engine;
 	gchar *word;
 
-	word = html_engine_get_word (e);
+	word = html_engine_get_spell_word (e);
 	if (word) {
 		spell_add_to_session (cd->html, word, cd);
 		g_free (word);
@@ -300,6 +355,11 @@ insert_html (GtkWidget *mi, GtkHTMLControlData *cd)
 		menuitem = gtk_menu_item_new_with_label (l); \
                 ADD_ITEM_BASE (f,t)
 
+#define ADD_ITEM_SENSITIVE(l,f,t,s) \
+		menuitem = gtk_menu_item_new_with_label (l); \
+                ADD_ITEM_BASE (f,t); \
+                gtk_widget_set_sensitive (menuitem, s);
+
 #define ADD_ITEM_UTF8(l,f,t) \
                 menuitem = e_utf8_gtk_menu_item_new_with_label (GTK_MENU (menu), l); \
                 ADD_ITEM_BASE (f,t)
@@ -338,6 +398,7 @@ prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
 	GtkWidget *submenu, *menuparent;
 	GtkWidget *menuitem;
 	guint items_sep = 0;
+	gboolean active = FALSE;
 
 	obj  = cd->html->engine->cursor->object;
 	menu = gtk_menu_new ();
@@ -353,11 +414,33 @@ prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
 	ADD_ITEM ("Insert HTML", insert_html, NONE);
 	ADD_SEP;
 #endif
+	active = html_engine_is_selection_active (e);
+	ADD_ITEM (_("Undo"), undo, NONE); 
+	ADD_ITEM (_("Redo"), redo, NONE); 
 
-	if (!html_engine_is_selection_active (e) && obj && html_object_is_text (obj) && !html_engine_word_is_valid (e)) {
+	ADD_SEP;
+	ADD_ITEM_SENSITIVE (_("Copy"), copy, NONE, active);
+	ADD_ITEM_SENSITIVE (_("Cut"),  cut, NONE, active);
+	ADD_ITEM (_("Paste"),  paste, NONE);
+	ADD_ITEM (_("Paste Quotation"),  paste_cite, NONE);
+
+	ADD_SEP;
+	ADD_ITEM (_("Insert link"), insert_link, NONE);
+	if (cd->format_html
+	    && ((active && html_engine_selection_contains_link (e))
+		|| (obj
+		    && (HTML_OBJECT_TYPE (obj) == HTML_TYPE_LINKTEXT
+			|| (HTML_OBJECT_TYPE (obj) == HTML_TYPE_IMAGE
+			    && (HTML_IMAGE (obj)->url
+				|| HTML_IMAGE (obj)->target)))))) {
+		ADD_ITEM (_("Remove link"), remove_link, NONE);
+	}
+
+	if (!active && obj && html_object_is_text (obj)
+	    && !html_engine_spell_word_is_valid (e)) {
 		gchar *spell, *word, *check_utf8, *add_utf8, *ignore_utf8, *ignore, *add;
 
-		word   = html_engine_get_word (e);
+		word   = html_engine_get_spell_word (e);
 		check_utf8 = e_utf8_from_locale_string (_("Check '%s' spelling..."));
 		spell  = g_strdup_printf (check_utf8, word);
 		g_free (check_utf8);
@@ -367,6 +450,7 @@ prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
 		ignore_utf8 = e_utf8_from_locale_string (_("Ignore '%s'"));
 		ignore = g_strdup_printf (ignore_utf8, word);
 		g_free (ignore_utf8);
+		ADD_SEP;
 		SUBMENU (N_("Spell checker"));
 		if (cd->has_spell_control) {
 			ADD_ITEM_UTF8 (spell, spell_check_cb, NONE);
@@ -383,37 +467,22 @@ prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
 		g_free (word);
 	}
 
-	if (cd->format_html && (html_engine_is_selection_active (e)
-				|| (obj
-				    && (HTML_OBJECT_TYPE (obj) == HTML_TYPE_LINKTEXT
-					|| (HTML_OBJECT_TYPE (obj) == HTML_TYPE_IMAGE
-					    && (HTML_IMAGE (obj)->url
-						|| HTML_IMAGE (obj)->target)))))) {
-		ADD_SEP;
-		ADD_ITEM (_("Remove link"), remove_link, NONE);
-	}
-
-	if (html_engine_is_selection_active (e)) {
-		ADD_SEP;
-		ADD_ITEM (_("Copy"), copy, NONE);
-		ADD_ITEM (_("Cut"),  cut, NONE);
-	}
-	if (e->clipboard) {
-		if (!html_engine_is_selection_active (e)) {
-			ADD_SEP;
-		}
-		ADD_ITEM (_("Paste"),  paste, NONE);
-	}
-	if (cd->format_html && html_engine_is_selection_active (e)) {
-		ADD_SEP;
-		ADD_ITEM (_("Text..."), prop_dialog, TEXT);
-		ADD_PROP (TEXT);
-		ADD_ITEM (_("Paragraph..."), prop_dialog, PARAGRAPH);
-		ADD_PROP (PARAGRAPH);
-		ADD_ITEM (_("Link..."), prop_dialog, LINK);
-		ADD_PROP (LINK);
-	} else if (cd->format_html && obj) {
+	if (cd->format_html && obj) {
 		switch (HTML_OBJECT_TYPE (obj)) {
+		case HTML_TYPE_TEXT:
+			ADD_SEP;
+			ADD_ITEM (_("Text..."), prop_dialog, TEXT);
+			ADD_PROP (TEXT);
+			ADD_ITEM (_("Paragraph..."), prop_dialog, PARAGRAPH);
+			ADD_PROP (PARAGRAPH);
+			break;
+		case HTML_TYPE_LINKTEXT:
+			ADD_SEP;
+			ADD_ITEM (_("Link..."), link_prop_dialog, LINK);
+			ADD_PROP (LINK);
+			ADD_ITEM (_("Paragraph..."), prop_dialog, PARAGRAPH);
+			ADD_PROP (PARAGRAPH);
+			break;
 		case HTML_TYPE_RULE:
 			ADD_SEP;
 			ADD_ITEM (_("Rule..."), prop_dialog, RULE);
@@ -425,23 +494,7 @@ prepare_properties_and_menu (GtkHTMLControlData *cd, guint *items)
 			ADD_PROP (IMAGE);
 			ADD_ITEM (_("Paragraph..."), prop_dialog, PARAGRAPH);
 			ADD_PROP (PARAGRAPH);
-			ADD_ITEM (_("Link..."), prop_dialog, LINK);
-			ADD_PROP (LINK);
 			break;
-		case HTML_TYPE_LINKTEXT:
-		case HTML_TYPE_TEXT:
-			ADD_SEP;
-			ADD_PROP (TEXT);
-			ADD_ITEM (_("Text..."), prop_dialog, TEXT);
-			ADD_PROP (PARAGRAPH);
-			ADD_ITEM (_("Paragraph..."), prop_dialog, PARAGRAPH);
-			ADD_PROP (LINK);
-			ADD_ITEM (_("Link..."), prop_dialog, LINK);
-			break;
-		case HTML_TYPE_TABLE:
-			ADD_SEP;
-			ADD_PROP (TABLE);
-			ADD_ITEM (_("Table..."), prop_dialog, TABLE);
 		default:
 		}
 		if (obj->parent && obj->parent->parent && HTML_IS_TABLE_CELL (obj->parent->parent)) {
@@ -498,10 +551,38 @@ popup_show (GtkHTMLControlData *cd, GdkEventButton *event)
 	return (items > 0);
 }
 
+static void
+set_position (GtkMenu *menu, gint *x, gint *y, gpointer data)
+{
+	GtkHTMLControlData *cd = (GtkHTMLControlData *) data;
+	HTMLEngine *e = cd->html->engine;
+	gint xw, yw;
+
+	gdk_window_get_origin (GTK_WIDGET (cd->html)->window, &xw, &yw);
+	html_object_get_cursor_base (e->cursor->object, e->painter, e->cursor->offset, x, y);
+	*x += xw + e->leftBorder;
+	*y += yw + e->topBorder;
+}
+
+gint
+popup_show_at_cursor (GtkHTMLControlData *cd)
+{
+	GtkWidget *menu;
+	guint items = 0;
+
+	menu = prepare_properties_and_menu (cd, &items);
+	gtk_widget_show (menu);
+	if (items)
+		gtk_menu_popup (GTK_MENU (menu), NULL, NULL, set_position, cd, 0, 0);
+	gtk_widget_unref (menu);
+
+	return (items > 0);
+}
+
 void
 property_dialog_show (GtkHTMLControlData *cd)
 {
-	guint items;
+	guint items = 0;
 
 	gtk_widget_unref (prepare_properties_and_menu (cd, &items));
 	if (items)
